#!/usr/bin/python

import rclpy
from rclpy.node import Node
from time import time
from std_msgs.msg import Int16
from geometry_msgs.msg import Twist


class Compass(Node):
    def __init__(self):
        super().__init__('compass')
        
        self.last = [0, 0, 0]

        self.publisher = self.create_publisher(Int16, 'compass', 10)
        self.sub = self.create_subscription(Int16, "modbus/compass", self.update, 10)
    
    def update(self, msg):
        self.last.append(msg.data)
        self.last = self.last[1:]

        a = self.last[0]
        b = self.last[1]
        c = self.last[2]
        msg.data = max(a, c) if (max(a, b) == max(b, c)) else max(b, min(a, c))
        # self.get_logger().info(f"{msg}")
        self.publisher.publish(msg)


def main():
    rclpy.init()
    compass  = Compass()
    rclpy.spin(compass)
    compass.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()