#!/usr/bin/python

import rclpy
from rclpy.node import Node
from time import time
from std_msgs.msg import Float64
from geometry_msgs.msg import Twist


class ToBerlin(Node):
    def __init__(self):
        super().__init__('to_berlin')
        self.last_time_joy = 0
        self.delta_time = 2.0

        self.publisher = self.create_publisher(Twist, 'twist_robot', 10)
        self.sub = self.create_subscription(Float64, "camera_front/compressed/z", self.update_front, 10)
        self.sub = self.create_subscription(Float64, "camera_side/compressed/z", self.update_side, 10)
        self.get_logger().info(f"Twist robot Ready")
    
    def update_front(self, num):
        self.last_time_joy = time()
        msg = Twist()

        msg.linear.x = 0.1
        msg.angular.z = 0.3 * num.data

        self.publisher.publish(msg)

    def update_side(self, num):
        if self.last_time_joy + self.delta_time < time():
            msg = Twist()

            msg.linear.x = 0.075
            msg.angular.z = -num.data

            self.publisher.publish(msg)



def main():
    rclpy.init()
    to_berlin = ToBerlin()
    rclpy.spin(to_berlin)
    to_berlin.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()