#!/usr/bin/python

import rclpy
from rclpy.node import Node
from time import time
from geometry_msgs.msg import Twist


class TwistController(Node):
    def __init__(self):
        super().__init__('twist_controller')

        self.last_time_joy = 0

        self.declare_parameter('delta_time', rclpy.Parameter.Type.DOUBLE)

        self.delta_time = self.get_declare_parameter("delta_time", lambda x: x > 0, "delta_time not more than zero")

        self.twist_joy =  self.create_subscription(Twist, "twist_joy", self.set_joy, 1)
        self.twist_robot =  self.create_subscription(Twist, "twist_robot", self.set_robot, 1)
        # self.create_timer(1 / 10, self.set_robot)

        self.publisher = self.create_publisher(Twist, 'twist_command', 10)

        self.get_logger().info('Twist controller started')
    
    def set_joy(self, msg):
        self.publisher.publish(msg)
        self.last_time_joy = time()
        # self.get_logger().info('pub Twist joy')

    def set_robot(self, *msg):
        if msg:
            if self.last_time_joy + self.delta_time < time():
                self.publisher.publish(msg[0])
                # self.get_logger().info('pub Twist robot')
              
    
    def get_declare_parameter(self, name, checker, error_text):
        value = self.get_parameter(name).value
        if not checker(value):
            self.get_logger().error(error_text)
            raise Exception(error_text)
        return value


def main():
    rclpy.init()
    twist_controller = TwistController()
    rclpy.spin(twist_controller)
    twist_controller.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()