#!/usr/bin/python

import rclpy
from rclpy.node import Node
from time import time
from std_msgs.msg import Float64, String, Int16
from geometry_msgs.msg import Point  # Point.x, Point.y, Point.z
from geometry_msgs.msg import Twist
from math import ceil, floor, radians, cos, sin
"""
ros2 topic pub -1 /ws11/modbus/calibration/write std_msgs/msg/Int16 "{data: 1}"
ros2 topic pub -1 /ws11/modbus/calibration/write std_msgs/msg/Int16 "{data: 0}"
ros2 topic echo /ws11/modbus/calibration
ros2 topic echo /ws11/modbus/compass
"""

class World(Node):
    is_twist = True
    move_type = "STOP"
    
    # ----- MISSIONS -----

    world_markers = {
        "start_point": [0, 0],
        "yellow_gate": [6, 0],
        "blue_gate": [11, 0],
        "blue_bou": [10, 0],
        "black_circle": [10, 10],
        "red_point": [10, 10],  #????
        "trash": [0, 10],
    }

    # start position
    position = [0, 0]
    heading = 0

    last_heading = 0
    last_heading_time = 0
    d_heads = [[0, 0] for _ in range(120)]
    
    __compass = 0
    # для бассейна
    compass_delta = 0

    # square_mission = {
    # #  point_name: [x, y] in meters
    #     "point_1": [1, 0],
    #     "point_2": [1, 1],
    #     "point_3": [0, 1],
    #     "point_4": [0, 0],
    # }

    # vlad24_mission = {
    # #  point_name: [x, y] in meters
    # # Прохождение ворот
    #     "point_1": [6, 0],
    #     # "tranzit": [8, -1],
    #     # "point_2": [9, 0],
    #     # "tranzit": [8, 1],
    #     # "point_1": [6, 0],
    # # Возвращение на старт
    #     "start":   [0, 0],
    # }

    # ----- SPEED -----

    zero_speed_signal = 1500
    # TODO посчитать
    # Скорость поворота в градусах за секунду при максимальной мощности
    rotation_speed = 180

    # Посчитанно Кириллом и Андреем в опытовом бассейне
    old_motor_speed = {
    # signal: speed (in m/s)
        1500: 0,
        1540: 0,
        1537: 0.10,
        1574: 0.17,
        1600: 0.40,
        1700: 0.73,
        1800: 0.95,
        1900: 1.04,
        2000: 1.04,
    }

    old_motor_ang_speed = {
        0: 0,
        # 0.15: -10,
        0.3: -20,
    }

    # Посчитанно Кириллом и Андреем в опытовом бассейне
    new_motor_speed = {
    # signal: speed (in m/s)
        1500: 0,
        1540: 0,
        1600: 0.50,
        1700: 0.97,
        1800: 1.15,
        1900: 1.29,
        2000: 1.29,
    }

    new_motor_ang_speed = {
        0: 0,
        # 0.15: -10,
        0.3: 0,
    }

    left_motor_signal = zero_speed_signal
    right_motor_signal = zero_speed_signal
    speed_signal = 0
    steering_signal = 0

    def update(self):
        delta_time = time() - self.last_time
        self.last_time = time()

        # Вычисление скорости катамарана
        if self.is_twist:
            med_signal = round(1_500 + abs(self.speed_signal) * 500)
        else:
            med_signal = (self.left_motor_signal + self.right_motor_signal) / 2
        low, high = 1500, 2000
        for key in self.motor_speed.keys():
            if low <= key <= med_signal:
                low = key
            if high >= key >= med_signal:
                high = key
                break

        speed = round(self.motor_speed[low] + ((med_signal - low) / (high - low) * (self.motor_speed[high] - self.motor_speed[low]) if high != low else 0), 2)

        # Расчёт виртуального курса
        if self.is_virtual_compas:
            if self.is_twist:
                delta_signals = self.steering_signal
                for key in self.motor_speed.keys():
                    if low <= key <= med_signal:
                        low = key
                    if high >= key >= med_signal:
                        high = key
                        break
            else:
                # TODO check md 1_000
                delta_signals = (self.right_motor_signal - self.left_motor_signal) / 2_000
            self.heading += delta_signals * self.rotation_speed * delta_time
        
        # Расчёт смещения
        """
        ось X направленна вверх (0 градусов)
        ось Y направленна направо (-90 градусов)
        """
        dx = speed *  cos(radians(self.heading)) * delta_time
        dy = speed *  sin(radians(self.heading)) * delta_time
        # Обновление позиции
        self.position[0] += dx
        self.position[1] += dy
        msg = Point()
        msg.x = float(self.position[0])
        msg.y = float(self.position[1])
        self.pub_position.publish(msg)

        d_head = self.heading - self.last_heading
        if d_head <= -180:
            d_head = 360 + d_head
        if d_head >= 180:
            d_head = d_head - 360
        d_time = time() - self.last_heading_time
        self.last_heading_time = time()
        self.last_heading = self.heading
        ang_speed = round((d_head) / (d_time))
        self.d_heads.append([ang_speed, d_time])
        self.d_heads.pop(0)
        ang_speed = round(sum(spd * t for spd, t in self.d_heads) / sum(t for spd, t in self.d_heads),1)

        self.get_logger().info(f"{ang_speed=} {round(d_head, 2)} {round(d_time, 2)}")
        # msg = Point()
        # msg.x = float(self.mission["point_1"][0])
        # msg.y = float(self.mission["point_1"][1])
        # self.pub_goal.publish(msg)
        # self.get_logger().info(f"POSITION INFO: {med_signal}:{speed}:{round(self.heading)}\t{round(dx, 2)}:{round(dy, 2)}:{round(delta_time, 2)}\t{round(self.position[0], 2)}:{round(self.position[1], 2)}")

    def __init__(self):
        super().__init__('world')

        # ----- PARAMETRS -----

        # "Угадывание" положения катамарана (очень примерно)
        if self.is_twist:
            # Для "угадывания" скорости катамарана (очень примерно)
            self.declare_parameter('twist_topic', rclpy.Parameter.Type.STRING)
        else:
            # Для "угадывания" скорости катамарана (очень примерно)
            self.declare_parameter('left_motor_signal', rclpy.Parameter.Type.STRING)
            self.declare_parameter('right_motor_signal', rclpy.Parameter.Type.STRING) 
            # Если надо инвертируем сигнал с мотора
            self.declare_parameter('is_invert_left', rclpy.Parameter.Type.INTEGER) 
            self.declare_parameter('is_invert_right', rclpy.Parameter.Type.INTEGER)

        # Тип кожуха мотора
        self.declare_parameter('is_new_motor', rclpy.Parameter.Type.INTEGER)
        # Компас (если пустое, то "угадываем" направление) TODO обработка реального сигнала с компаса
        self.last_compass = []
        self.declare_parameter('compas_topic', rclpy.Parameter.Type.STRING)
        self.declare_parameter('heading_topic', rclpy.Parameter.Type.STRING)
        # hamach hz in main loop
        self.declare_parameter('hz', rclpy.Parameter.Type.INTEGER)
        # Топик обнуление координат
        self.declare_parameter('reset_topic', rclpy.Parameter.Type.STRING)
        # Топик координат
        self.declare_parameter('position_topic', rclpy.Parameter.Type.STRING)
        # self.declare_parameter('goal_topic', rclpy.Parameter.Type.STRING)
        # Топики найденных объектов
        world_markers_topics_parametrs = {
            "gate_position": "yellow_gate",
            "blue_position": "blue_bou",
            "blue_gate_position": "blue_gate",
            # "black_position": "black_circle",
            # "red_position": "red_point",
        }
        for key in world_markers_topics_parametrs.keys():
            self.declare_parameter(key, rclpy.Parameter.Type.STRING)


        # Миссия
        # TODO Выполнение миссии
        # self.declare_parameter('mission_name', rclpy.Parameter.Type.STRING)

        # ----- VALIDATION AND SETTING -----
        if self.is_twist:
            # Подписываемся на twist
            self.create_subscription(Twist, self.get_declare_parameter('twist_topic', lambda x: x, "no twist_topic is described"), self.twist_update, 1)
        else:
            # Если надо, то будем инвертировать сигнал (если программно фиксим неправильное подключение обмоток двигателя)
            self.is_invert_left = self.get_declare_parameter('is_invert_left', lambda x: x in [0, 1], "is_new_motor can be 0 or 1") == 1
            self.is_invert_right = self.get_declare_parameter('is_invert_right', lambda x: x in [0, 1], "is_new_motor can be 0 or 1") == 1
            # Подписываемся на сигнал от модбаса
            self.create_subscription(Float64, self.get_declare_parameter('left_motor_signal', lambda x: x, "no left_motor_signal is described"), self.left_motor_update, 1)
            self.create_subscription(Float64, self.get_declare_parameter('right_motor_signal', lambda x: x, "no right_motor_signal is described"), self.right_motor_update, 1)

        # Имеется в виду новая насадка на мотор или нет (новая - от Кирилла (обтекаемая) или старая - от трионикса (белая))
        if self.get_declare_parameter('is_new_motor', lambda x: x in [0, 1], "is_new_motor can be 0 or 1"):
            self.motor_speed = self.new_motor_speed
            self.motor_ang_speed = self.new_motor_ang_speed
        else:
            self.motor_speed = self.old_motor_speed
            self.motor_ang_speed = self.old_motor_ang_speed

        # Выбор миссии
        # mission_name = self.get_declare_parameter('mission_name', lambda x: x in ["vlad24", "square"], "mission not found")
        # if mission_name == "vlad24":
        #     self.mission = self.vlad24_mission
        # elif mission_name == "square":
        #     self.mission = self.square_mission
        
        # Подписываемся на найденные объекты
        self.create_subscription(String, "move_type", self.move_type_callback, 10)
        for key, value in world_markers_topics_parametrs.items():
            self.create_subscription(Point, self.get_declare_parameter(key, lambda x: x, f"no {key} topic is described"), self.get_marker_callback(value), 1)


        # Если топик компаса не указан, то включается режим виртуального компаса
        compas_topic = self.get_declare_parameter('compas_topic', lambda x: x, "no compas topic")
        if compas_topic:
            self.create_subscription(Int16, compas_topic, self.compas_update, 1)
            self.is_virtual_compas = False
        else:
            self.is_virtual_compas = True
        self.pub_heading = self.create_publisher(Int16, self.get_declare_parameter('heading_topic', lambda x: x, "no heading_topic is described"), 10)


        self.create_subscription(String, self.get_declare_parameter('reset_topic', lambda x: x, "no reset_topic is described"), self.reset, 10)
        self.pub_position = self.create_publisher(Point, self.get_declare_parameter('position_topic', lambda x: x, "no position_topic is described"), 10)
        # self.pub_goal = self.create_publisher(Point, self.get_declare_parameter('goal_topic', lambda x: x, "no goal_topic is described"), 10)

        # main loop
        self.last_time = time()
        self.create_timer(1 / self.get_declare_parameter('hz', lambda x: x > 0, "hz is less or equal zero"), self.update)
        if self.is_twist:
            self.get_logger().info(f"initial params: {self.is_virtual_compas = } {compas_topic = }")
        else:
            self.get_logger().info(f"initial params: {self.is_invert_left = } {self.is_invert_right = } {self.is_virtual_compas = } {compas_topic = }")
    
    def twist_update(self, msg):
        self.speed_signal = msg.linear.x
        self.steering_signal = msg.angular.z

    def left_motor_update(self, msg):
        signal = msg.data

        if self.is_invert_left:
            signal = -(signal - self.zero_speed_signal) + self.zero_speed_signal
        
        self.left_motor_signal = signal

    def right_motor_update(self, msg):
        signal = msg.data

        if self.is_invert_left:
            signal = -(signal - self.zero_speed_signal) + self.zero_speed_signal
        
        self.right_motor_signal = signal

    def reset(self, msg):
        if msg.data == "reset_map":
            self.get_logger().info(f"reset map")
            self.position = [0, 0]
            if self.is_virtual_compas:
                self.get_logger().info(f"reset virtual compass")
                self.heading = 0
                self.heading = self.heading
        elif msg.data == "reset_heading":
            self.get_logger().info(f"reset heading {self.__compass}")
            self.compass_delta = self.__compass

    def get_marker_callback(self, name):
        def callback(msg):
            self.update_position(msg, name)
        return callback

    def update_position(self, msg, name):
        if self.move_type == "MOVE TO" and name in ["yellow_gate", "blue_gate"]:
            obj_pos = self.world_markers[name]
            self.position = [obj_pos[0] - msg.x, obj_pos[1] - msg.y]
        # if self.move_type == "MOVE CIRCLE" and name == "blue_bou":
        #     obj_pos = self.world_markers[name]
        #     self.position = [obj_pos[0] - msg.x, obj_pos[1] - msg.y]

    def compas_update(self, msg):
        # Проверка на правильность преобразования
        assert -10%360 == 350

        self.__compass = msg.data
        self.last_compass.append((msg.data - self.compass_delta) % 360)
        self.last_compass = self.last_compass[1:]

        try:
            a = self.last_compass[0]
            b = self.last_compass[1]
            c = self.last_compass[2]
        except IndexError:
            self.last_compass.append(msg.data)
            self.last_compass.append(msg.data)
            self.last_compass.append(msg.data)
            a = self.last_compass[0]
            b = self.last_compass[1]
            c = self.last_compass[2]
        
        self.heading = int(max(a, c) if (max(a, b) == max(b, c)) else max(b, min(a, c)))

        # self.get_logger().info(f"initial params: {self.heading} {self.compass_delta}")
        self.pub(self.pub_heading, Int16, {"data": self.heading})
    
    def get_declare_parameter(self, name, checker, error_text):
        value = self.get_parameter(name).value
        if not checker(value):
            self.get_logger().error(error_text)
            raise Exception(error_text)
        return value
    
    def pub(self, topic, data_type, data: dict):
        msg = data_type()
        for key, value in data.items():
            setattr(msg, key, value)
        topic.publish(msg)
    
    def move_type_callback(self, msg):
        self.move_type = msg.data



def main():
    rclpy.init()
    world = World()
    rclpy.spin(world)
    world.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
